
from .info import check_PY2
check_PY2()

import __builtin__ as builtins
import sys
from ctypes import pythonapi, c_char_p, c_void_p, py_object

from .streams import STDIN, STDOUT
from .readline_hook import check_encodings, stdio_readline


original_raw_input = builtins.raw_input
original_input = builtins.input

RETURN_UNICODE = True


PyOS_Readline = pythonapi.PyOS_Readline
PyOS_Readline.restype = c_char_p
PyOS_Readline.argtypes = [c_void_p, c_void_p, c_char_p]

PyFile_AsFile = pythonapi.PyFile_AsFile
PyFile_AsFile.restype = c_void_p
PyFile_AsFile.argtypes = [py_object]

STDIN_FILE_POINTER = PyFile_AsFile(sys.stdin)
STDOUT_FILE_POINTER = PyFile_AsFile(sys.stdout)


def stdout_encode(s):
	if isinstance(s, bytes):
		return s
	encoding = sys.stdout.encoding
	errors = sys.stdout.errors
	if errors is not None:
		return s.encode(encoding, errors)
	else:
		return s.encode(encoding)

def stdin_encode(s):
	if isinstance(s, bytes):
		return s
	encoding = sys.stdin.encoding
	errors = sys.stdin.errors
	if errors is not None:
		return s.encode(encoding, errors)
	else:
		return s.encode(encoding)

def stdin_decode(b):
	if isinstance(b, unicode):
		return b
	encoding = sys.stdin.encoding
	errors = sys.stdin.errors
	if errors is not None:
		return b.decode(encoding, errors)
	else:
		return b.decode(encoding)

def readline(prompt=""):
	check_encodings()
	prompt_bytes = stdout_encode(prompt)
	line_bytes = PyOS_Readline(STDIN_FILE_POINTER, STDOUT_FILE_POINTER, prompt_bytes)
	if line_bytes is None:
		raise KeyboardInterrupt
	else:
		return line_bytes


def raw_input(prompt=""):
	"""raw_input([prompt]) -> string

Read a string from standard input.  The trailing newline is stripped.
If the user hits EOF (Unix: Ctl-D, Windows: Ctl-Z+Return), raise EOFError.
On Unix, GNU readline is used if enabled.  The prompt string, if given,
is printed without a trailing newline before reading."""
	
	sys.stderr.flush()
	
	tty = STDIN.is_a_TTY() and STDOUT.is_a_TTY()
	
	if RETURN_UNICODE:
		if tty:
			line_bytes = readline(prompt)
			line = stdin_decode(line_bytes)
		else:
			line = stdio_readline(prompt)
		
	else:
		if tty:
			line = readline(prompt)
		else:
			line_unicode = stdio_readline(prompt)
			line = stdin_encode(line_unicode)
	
	if line:
		return line[:-1] # strip strailing "\n"
	else:
		raise EOFError

def input(prompt=""):
	"""input([prompt]) -> value

Equivalent to eval(raw_input(prompt))."""
	
	string = stdin_decode(raw_input(prompt))
	
	caller_frame = sys._getframe(1)
	globals = caller_frame.f_globals
	locals = caller_frame.f_locals
	
	return eval(string, globals, locals)


def enable(return_unicode=RETURN_UNICODE):
	global RETURN_UNICODE
	RETURN_UNICODE = return_unicode
	
	builtins.raw_input = raw_input
	builtins.input = input

def disable():
	builtins.raw_input = original_raw_input
	builtins.input = original_input




from __future__ import print_function # PY2

import __main__
import code
import sys

from .info import PY2


def print_banner(file=sys.stderr):
	print("Python {} on {}".format(sys.version, sys.platform), file=file)
	print('Type "help", "copyright", "credits" or "license" for more information.', file=file)

# PY3 # class InteractiveConsole(code.InteractiveConsole):
class InteractiveConsole(code.InteractiveConsole, object):
	# code.InteractiveConsole without banner
	# exits on EOF
	# also more robust treating of sys.ps1, sys.ps2
	# prints prompt into stderr rather than stdout
	# flushes sys.stderr and sys.stdout
	
	def __init__(self, locals=None, filename="<stdin>"):
		self.done = False
		# PY3 # super().__init__(locals, filename)
		super(InteractiveConsole, self).__init__(locals, filename)
	
	def raw_input(self, prompt=""):
		sys.stderr.write(prompt)
		if PY2:
			return raw_input()
		else:
			return input()
	
	def runcode(self, code):
		# PY3 # super().runcode(code)
		super(InteractiveConsole, self).runcode(code)
		sys.stderr.flush()
		sys.stdout.flush()
	
	def interact(self):
		#sys.ps1 = "~>> "
		#sys.ps2 = "~.. "
		
		try:
			sys.ps1
		except AttributeError:
			sys.ps1 = ">>> "
		
		try:
			sys.ps2
		except AttributeError:
			sys.ps2 = "... "
		
		more = 0
		while not self.done:
			try:
				if more:
					try:
						prompt = sys.ps2
					except AttributeError:
						prompt = ""
				else:
					try:
						prompt = sys.ps1
					except AttributeError:
						prompt = ""
				
				try:
					line = self.raw_input(prompt)
				except EOFError:
					self.on_EOF()
				else:
					more = self.push(line)
				
			except KeyboardInterrupt:
				self.write("\nKeyboardInterrupt\n")
				self.resetbuffer()
				more = 0
	
	def on_EOF(self):
		self.write("\n")
		# PY3 # raise SystemExit from None
		raise SystemExit


running_console = None

def enable():
	global running_console
	
	if running_console is not None:
		raise RuntimeError("interactive console already running")
	else:
		running_console = InteractiveConsole(__main__.__dict__) 
		running_console.interact() 

def disable():
	global running_console
	
	if running_console is None:
		raise RuntimeError("interactive console is not running")
	else:
		running_console.done = True
		running_console = None





import ctypes
from ctypes import (byref, POINTER, pythonapi, 
	c_int, c_char, c_char_p, c_void_p, py_object, c_ssize_t)

from .info import PY2


c_ssize_p = POINTER(c_ssize_t)

PyObject_GetBuffer = pythonapi.PyObject_GetBuffer
PyBuffer_Release = pythonapi.PyBuffer_Release


PyBUF_SIMPLE = 0
PyBUF_WRITABLE = 1


class Py_buffer(ctypes.Structure):
	_fields_ = [
		("buf", c_void_p),
		("obj", py_object),
		("len", c_ssize_t),
		("itemsize", c_ssize_t),
		("readonly", c_int),
		("ndim", c_int),
		("format", c_char_p),
		("shape", c_ssize_p),
		("strides", c_ssize_p),
		("suboffsets", c_ssize_p),
		("internal", c_void_p)
	]
	
	if PY2:
		_fields_.insert(-1, ("smalltable", c_ssize_t * 2))
	
	@classmethod
	def get_from(cls, obj, flags=PyBUF_SIMPLE):
		buf = cls()
		PyObject_GetBuffer(py_object(obj), byref(buf), flags)
		return buf
	
	def release(self):
		PyBuffer_Release(byref(self))


def get_buffer(obj, writable=False):
	buf = Py_buffer.get_from(obj, PyBUF_WRITABLE if writable else PyBUF_SIMPLE)
	try:
		buffer_type = c_char * buf.len
		return buffer_type.from_address(buf.buf)
	finally:
		buf.release()

